"""A collection of stand-alone functions for plotting purposes."""

from typing import Any, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

MATLAB_COLORS = [
    "#0072BD",
    "#D95319",
    "#EDB120",
    "#7E2F8E",
    "#77AC30",
    "#4DBEEE",
    "#A2142F",
]


def save2tikz(*figs) -> None:
    """Saves the figure to a tikz file (``.tex`` extension). See
    `tikzplotlib <https://pypi.org/project/tikzplotlib/>`_ for more details.

    Parameters
    ----------
    figs : :class:`matplotlib.figure.Figure`
        One or more matplotlib figures to be converted to tikz files. These
        files will be named based on the number of the corresponding figure.
    """
    import matplotlib as mpl
    import tikzplotlib

    # monkey patching to fix some issues with tikzplotlib
    mpl.lines.Line2D._us_dashSeq = property(lambda self: self._dash_pattern[1])
    mpl.lines.Line2D._us_dashOffset = property(lambda self: self._dash_pattern[0])
    mpl.legend.Legend._ncol = property(lambda self: self._ncols)
    for fig in figs:
        tikzplotlib.save(
            f"figure_{fig.number}.tex",
            figure=fig,
            extra_axis_parameters={r"tick scale binop=\times"},
        )


def spy(
    H: Union[cs.SX, cs.MX, cs.DM, cs.Sparsity, npt.ArrayLike],
    ax=None,
    **spy_kwargs: Any,
):
    """Implementation equivalent to :func:`matplotlib.pyplot.spy` that works also with
    CasADi symbolic matrices.

    Parameters
    ----------
    H : casadi SX, MX, DM, Sparsity or array_like
        The matrix (or its sparsity pattern) to spy.
    ax : :class:`matplotlib.axes.Axes`, optional
        The axis to draw the result on. If ``None``, creates a new axis.
    spy_kwargs
        Other arguments passed directly to :func:`matplotlib.pyplot.spy`.

    Returns
    -------
    :class:`matplotlib.image.AxesImage` or :class:`matplotlib.lines.Line2D`
        Same return types of :func:`matplotlib.pyplot.spy`.
    """
    import matplotlib.pyplot as plt

    if hasattr(H, "sparsity") and callable(H.sparsity):
        H = H.sparsity()
    if hasattr(H, "spy") and callable(H.spy):
        from contextlib import redirect_stdout
        from io import StringIO

        f = StringIO()
        with redirect_stdout(f):
            H.spy()
        out = f.getvalue()
        H = np.asarray(
            [
                list(line)
                for line in out.replace(".", "0").replace("*", "1").splitlines()
            ],
            dtype=int,
        )
    else:
        H = np.asarray(H)

    if ax is None:
        _, ax = plt.subplots(1, 1)
    o = ax.spy(H, **spy_kwargs)
    nz = np.count_nonzero(H)
    ax.set_xlabel(f"nz = {nz} ({nz / H.size * 100:.2f}%)")
    return o


def set_mpl_defaults(
    mpl_style: str = "bmh",
    linewidth: float = 1.5,
    markersize: float = 2,
    savefig_dpi: int = 600,
    np_print_precision: int = 4,
    matlab_colors: bool = False,
) -> None:
    """Sets some default parameters for :mod:`numpy` and :mod:`matplotlib` for printing
    and plotting.

    Parameters
    ----------
    mpl_style : str, optional
        :mod:`matplotlib` plotting style, by default ``"bmh"``.
    linewidth : float, optional
        :mod:`matplotlib` default linewidth, by default ``1.5``.
    markersize : float, optional
        :mod:`matplotlib` default markersize, by default ``2``.
    savefig_dpi : int, optional
        :mod:`matplotlib` savefig dpi, by default ``600``.
    np_print_precision : int, optional
        :mod:`numpy` printing precision, by default ``4``.
    matlab_colors : bool, optional
        Whether :mod:`matplotlib` should use Matlab colors for plotting, by default
        ``False``.
    """
    import matplotlib as mpl

    np.set_printoptions(precision=np_print_precision)
    mpl.style.use(mpl_style)  # 'seaborn-darkgrid'
    mpl.rcParams["lines.solid_capstyle"] = "round"
    mpl.rcParams["lines.linewidth"] = linewidth
    mpl.rcParams["lines.markersize"] = markersize
    mpl.rcParams["savefig.dpi"] = savefig_dpi
    if matlab_colors:
        from cycler import cycler

        mpl.rcParams["axes.prop_cycle"] = cycler("color", MATLAB_COLORS)



"""A collection of classes to perform scaling of variables in an :class:`csnlp.Nlp`
instance wrapped with :class:`csnlp.wrappers.NlpScaling`. The classes in this module
inform the wrapper on which variables or parameters to scale and how to scale them."""

from typing import Optional

import numpy as np
import numpy.typing as npt


class Scaler(dict[str, tuple[npt.ArrayLike, npt.ArrayLike]]):
    r"""Class for scaling of NLP variables. It suffices to register the scale of each
    the variable, and then it can be easily (un-)scaled according to

    .. math:: x_{scaled} = \frac{x - loc}{scale}.

    Quantities need not be numerical, and can be of any type that supports basic
    algebraic operations.

    Parameters
    ----------
    d : dict of (str, tuple of 2 array_like), optional
        A possible non-empty dict of variable group names with the corresponding scaling
        parameters ``loc`` and ``scale``.
    """

    def __init__(
        self, d: Optional[dict[str, tuple[npt.ArrayLike, npt.ArrayLike]]] = None
    ) -> None:
        super().__init__()
        if d is not None:
            for k, v in d.items():
                self.register(k, *v)

    def register(
        self, name: str, loc: npt.ArrayLike = 0, scale: npt.ArrayLike = 1
    ) -> None:
        """Registers for the variable ``name`` a new set of ``loc`` and ``scale``
        parameters for its normalization, but raises if duplicates occur.

        Parameters
        ----------
        name : str
            Variable group to register for normalization.
        loc : array_like or supports_algebraic_operations, optional
            Mean value of the variable, by default ``0``.
        scale : array_like or supports_algebraic_operations, optional
            Standard deviation of the variable, by default ``1``.

        Raises
        ------
        KeyError
            Raises if ``name`` is duplicated.
        """
        if name in self:
            raise KeyError(f"'{name}' already registered for normalization.")
        self[name] = (loc, scale)

    def can_scale(self, name: str) -> bool:
        """Checks whether the given variable's name was previously registered for
        scaling.

        Parameters
        ----------
        name : str
            Name of the variable to scale.

        Returns
        -------
        bool
            Whether the variable ``name`` has been registered.
        """
        return name in self

    def scale(self, name: str, x: npt.ArrayLike) -> npt.ArrayLike:
        """Scales the value ``x`` according to the ranges of variable ``name``.

        Parameters
        ----------
        name : str
            Variable group ``x`` belongs to.
        x : array_like or supports_algebraic_operations
            Value to be scaled.

        Returns
        -------
        array_like or supports_algebraic_operations
            Normalized ``x``.

        Raises
        ------
        AssertionError
            Raises if the scaled output's shape does not match input's shape.
        """
        loc, scale = self[name]
        out = (x - loc) / scale  # type: ignore[operator]
        assert np.shape(out) == np.shape(x), "Scaling altered input shape."
        return out

    def unscale(self, name: str, x: npt.ArrayLike) -> npt.ArrayLike:
        """Unscales the value ``x`` according to the ranges of ``name``.

        Parameters
        ----------
        name : str
            Variable group ``x`` belongs to.
        x : array_like or supports_algebraic_operations
            Value to be unscaled.

        Returns
        -------
        array_like
            Denormalized ``x``.

        Raises
        ------
        AssertionError
            Raises if the unscaled output's shape does not match input's shape.
        """
        loc, scale = self[name]
        out = x * scale + loc  # type: ignore[operator]
        assert np.shape(out) == np.shape(x), "Unscaling altered input shape."
        return out

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {super().__repr__()}>"

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {super().__str__()}>"


class MinMaxScaler(Scaler):
    r"""Class for scaling of quantities. It suffices to register the scale of each the
    variable, and then it can be easily (un-)scaled according to

    .. math:: x_{scaled} = \frac{x - min}{max - min}.

    Quantities need not be numerical, and can be of any type that supports basic
    algebraic operations.
    """

    def register(
        self, name: str, min: npt.ArrayLike = 0, max: npt.ArrayLike = 1
    ) -> None:
        """Registers for the variable ``name`` a new set of ``min`` and ``max``
        parameters for its normalization, but raises if duplicates occur.

        Parameters
        ----------
        name : str
            Variable group to register for normalization.
        min : array_like or supports_algebraic_operations, optional
            Minimum of the variable, by default ``0``.
        max : array_like or supports_algebraic_operations, optional
            Maximum of the variable, by default ``1``.

        Raises
        ------
        KeyError
            Raises if ``name`` is duplicated.
        """
        return super().register(name, min, max - min)



import warnings
from collections.abc import Iterator, Sequence
from itertools import count
from typing import Any, ClassVar, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from joblib import Memory

from ..core.debug import NlpDebug
from ..util.io import SupportsDeepcopyAndPickle
from .objective import HasObjective

SymType = TypeVar("SymType", cs.SX, cs.MX)


class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
    r"""The generic NLP class is a controller that solves a (possibly, nonlinear)
    optimization problem to yield a (possibly, sub-) optimal solution. This is a generic
    implementation in the sense that it does not solve a particular problem, but only
    offers the generic methods to build one (e.g., variables, constraints, objective,
    solver).

    Parameters
    ----------
    sym_type : {"SX", "MX"}, optional
        The CasADi symbolic variable type to use in the NLP, by default ``"SX"``.
    remove_redundant_x_bounds : bool, optional
        If ``True``, then redundant entries in :meth:`lbx` and :meth:`ubx` are removed
        when properties :meth:`h_lbx` and :meth:`h_ubx` are called. See these two
        properties for more details. By default, ``True``.
    cache : joblib.Memory, optional
        Optional cache to avoid computing the same exact NLP more than once. By default,
        no caching occurs.
    name : str, optional
        Name of the NLP scheme. If `None`, it is automatically assigned.
    debug : bool, optional
        If ``True``, the NLP logs in the :meth:`debug` property information regarding
        the creation of parameters, variables and constraints. By default, ``False``.

    Raises
    ------
    AttributeError
        Raises if the specified CasADi's symbolic type is neither ``"SX"`` nor ``"MX"``.

    Notes
    -----
    Constraints are internally handled in their canonical form, i.e., :math:`g(x,p) = 0`
    and :math:`h(x,p) \leq 0`. The objective :math:`f(x,p)` is always a scalar function
    to be minimized.
    """

    __ids: ClassVar[Iterator[int]] = count(0)
    is_multi: ClassVar[bool] = False
    """Flag to indicate that this is not a multistart NLP."""

    def __init__(
        self,
        sym_type: Literal["SX", "MX"] = "SX",
        remove_redundant_x_bounds: bool = True,
        cache: Memory = None,
        name: Optional[str] = None,
        debug: bool = False,
    ) -> None:
        id = next(self.__ids)
        name = f"{self.__class__.__name__}{id}" if name is None else name
        HasObjective.__init__(self, sym_type, remove_redundant_x_bounds, cache, name)
        SupportsDeepcopyAndPickle.__init__(self)
        self.id = id
        self._debug = NlpDebug() if debug else None

    @property
    def sym_type(self) -> type[SymType]:
        """Gets the CasADi symbolic type used in this NLP scheme."""
        return self._sym_type

    @property
    def debug(self) -> Optional[NlpDebug]:
        """Gets debug information on the NLP scheme."""
        return self._debug

    @property
    def unwrapped(self) -> "Nlp[SymType]":
        """Returns the original NLP of the wrapper."""
        return self

    def is_wrapped(self, *args: Any, **kwargs: Any) -> bool:
        """Gets whether the NLP instance is wrapped or not by the given wrapper type."""
        return False

    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        out = super().parameter(name, shape)
        if self._debug is not None:
            self._debug.register("p", name, shape)
        return out

    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        out = super().variable(name, shape, discrete, lb, ub)
        if self._debug is not None:
            self._debug.register("x", name, shape)
        return out

    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        out = super().constraint(name, lhs, op, rhs, soft, simplify)
        if self._debug is not None:
            self._debug.register("g" if op == "==" else "h", name, out[0].shape)
        return out

    def to_function(
        self,
        name: str,
        ins: Sequence[SymType],
        outs: Sequence[SymType],
        name_in: Optional[Sequence[str]] = None,
        name_out: Optional[Sequence[str]] = None,
        opts: Optional[dict[Any, Any]] = None,
    ) -> cs.Function:
        """Converts the optimization problem to an ``MX`` symbolic
        :class:`casadi.Function`. If the NLP is modelled in ``SX``, the function will
        still be converted to ``MX`` since the IPOPT interface cannot expand ``SX``
        for now.

        Parameters
        ----------
        name : str
            Name of the function.
        ins : sequence of casadi.SX or MX
            Input variables of the function. These must be expressions
            providing the parameters of the NLP and the initial conditions of
            the primal variables ``x``.
        outs : sequence of casadi.SX or MX
            Output variables of the function. These must be expressions
            depending on the primal variable ``x``, parameters ``p``, and dual
            variables ``lam_g``, ``lam_h``, ``lam_lbx``, ``lam_ubx`` of the NLP.
        name_in : sequence of str, optional
            Name of the inputs, by default None.
        name_out : sequence of str, optional
            Name of the outpus, by default None.
        opts : dict[Any, Any], optional
            Options to be passed to :class:`casadi.Function`, by default None.

        Returns
        -------
        casadi.Function
            The NLP solver as an instance of :class:`casadi.Function`.

        Raises
        ------
        RuntimeError
            Raises if the solver is uninitialized, or if the input or output expressions
            have free variables that are not provided or cannot be computed by the
            solver.
        """
        if self._sym_type is cs.SX:
            warnings.warn(
                "The IPOPT interface does not support SX expansion, "
                "so the function must be wrapped in MX.",
                RuntimeWarning,
            )
        S = self._solver
        if S is None:
            raise RuntimeError("Solver not yet initialized.")

        # converts inputs/outputs to/from variables and parameters
        n_outs = len(outs)
        Fin = cs.Function("Fin", ins, [self._x, self._p])
        Fout = cs.Function(
            "Fout",
            [self._p, self._x, self._lam_g, self._lam_h, self._lam_lbx, self._lam_ubx],
            outs,
        )

        # call the solver
        if self._sym_type is cs.SX:
            Fin = Fin.wrap()
            Fout = Fout.wrap()
            ins = [Fin.mx_in(i) for i in range(len(ins))]
            outs = [Fout.mx_out(i) for i in range(len(outs))]
        x0, p = Fin(*ins)
        sol = S(
            x0=x0,
            p=p,
            lbx=self._lbx.data,
            ubx=self._ubx.data,
            lbg=np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            ubg=0,
            lam_x0=0,
            lam_g0=0,
        )
        x = sol["x"]
        lam_g = sol["lam_g"][: self.ng, :]
        lam_h = sol["lam_g"][self.ng :, :]
        lam_lbx = -cs.fmin(sol["lam_x"], 0)[self.nonmasked_lbx_idx, :]
        lam_ubx = cs.fmax(sol["lam_x"], 0)[self.nonmasked_ubx_idx, :]
        Fsol = cs.Function("Fsol", ins, [x, lam_g, lam_h, lam_lbx, lam_ubx])

        # build final function
        final_outs = Fout(p, *Fsol(*ins))
        if n_outs == 1:
            final_outs = [final_outs]
        args = [name, ins, final_outs]
        if name_in is not None and name_out is not None:
            args.extend((name_in, name_out))
        if opts is not None:
            args.append(opts)
        return cs.Function(*args)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return super().solve(*args, **kwargs)

    def __str__(self) -> str:
        """Returns the NLP name and a short description."""
        msg = "not initialized" if self._solver is None else "initialized"
        C = len(self._cons)
        return (
            f"{type(self).__name__} {{\n"
            f"  name: {self.name}\n"
            f"  #variables: {len(self._vars)} (nx={self.nx})\n"
            f"  #parameters: {len(self._pars)} (np={self.np})\n"
            f"  #constraints: {C} (ng={self.ng}, nh={self.nh})\n"
            f"  CasADi solver {msg}.\n}}"
        )

    def __repr__(self) -> str:
        """Returns the string representation of the NLP instance."""
        return f"{type(self).__name__}: {self.name}"



"""A module that provides parallelization of optimization problems for multistarting.

Motivation
==========

Sometimes, NLP problems will converge to poor local minima. This is especially true
when the nonlinearities in the problem are strong and nonconvex and the number of
optimization variables is high, increasing the chances of the gradient-based solver
getting stuck in a local minimum.

One way to mitigate this issue is to use multistart methods, which consist in running
the optimization problem from multiple starting points (i.e., initial guesses for the
primal variables) and selecting the best solution out of the ones found. The benefit is
straightforward: the more starting points are used and the more spread these are, the
higher the chances of finding a better solution (possibly, but difficulty, the global
minimum). The downside is that the computational cost increases with the number of
starting points, in particular when the repeated optimization problems are run serially.
To overcome this, parallelization of the computations is an intuitive yet effective
solution.

Overview
========

To address this problem, the :mod:`csnlp.multistart` module provides classes to run
multistart optimization problems in parallel, with different implementations of the
parallelization. These classes maintain the same interface from :class:`csnlp.Nlp`, so
from the perspective of the user little effort is required to employ multistart. In
particular, the module provides the following classes:

- :class:`MappedMultistartNlp` runs the optimization problems in parallel
  using the :func:`casadi.Function.map` function
- :class:`ParallelMultistartNlp` runs the optimization problems in parallel using the
  :class:`joblib.Parallel` class
- :class:`StackedMultistartNlp` runs the optimization problems in parallel by stacking
  them multiple times in a single large-scale optimization problem.

These are all subclasses of the base class :class:`MultistartNlp`, which provides the
interface for multistarting optimization problems (see
:meth:`MultistartNlp.solve_multi`). As aforementioned, the effectiveness of multistart
methods depends also on the quality and spread of the starting points. This module thus
also provides classes to generate these points in a structured or random way:

- :class:`RandomStartPoint` and :class:`RandomStartPoints` allows to generate random
  starting points and group them, respectively
- :class:`StructuredStartPoint` and :class:`StructuredStartPoints` allows to generate
  structured (i.e., deterministic) starting points and group them, respectively.
"""

__all__ = [
    "MappedMultistartNlp",
    "MultistartNlp",
    "ParallelMultistartNlp",
    "RandomStartPoint",
    "RandomStartPoints",
    "StackedMultistartNlp",
    "StructuredStartPoint",
    "StructuredStartPoints",
]

from .multistart_nlp import (
    MappedMultistartNlp,
    MultistartNlp,
    ParallelMultistartNlp,
    StackedMultistartNlp,
)
from .startpoints import (
    RandomStartPoint,
    RandomStartPoints,
    StructuredStartPoint,
    StructuredStartPoints,
)



from functools import cached_property
from typing import Literal, TypeVar

import casadi as cs
import numpy as np
from numpy.typing import NDArray

from ..core.cache import invalidate_cache
from .parameters import HasParameters

SymType = TypeVar("SymType", cs.SX, cs.MX)


class HasVariables(HasParameters[SymType]):
    """Class for the creation and storage of symbolic variables in an NLP problem. It
    builds on top of :class:`HasParameters`, which handles parameters.

    Parameters
    ----------
    sym_type : {"SX", "MX"}, optional
        The CasADi symbolic variable type to use in the NLP, by default ``"SX"``.
    """

    def __init__(self, sym_type: Literal["SX", "MX"] = "SX") -> None:
        super().__init__(sym_type)
        self._vars: dict[str, SymType] = {}
        self._x = self._sym_type()
        self._has_discrete = False
        self._discrete: dict[str, bool] = {}

    @property
    def x(self) -> SymType:
        """Gets the primary variables of the NLP scheme in vector form."""
        return self._x

    @property
    def nx(self) -> int:
        """Number of variables in the NLP scheme."""
        return self._x.shape[0]

    @property
    def variables(self) -> dict[str, SymType]:
        """Gets the primal variables of the NLP scheme."""
        return self._vars

    @property
    def has_discrete(self) -> bool:
        """Flags if the NLP has discrete variables."""
        return self._has_discrete

    @cached_property
    def discrete(self) -> NDArray[np.bool_]:
        """Gets the boolean array indicating which variables are discrete."""
        if not self._has_discrete:
            return np.zeros(self.nx, dtype=bool)
        vars = self._vars
        return np.concatenate(
            [
                (np.ones if is_discrete else np.zeros)(
                    np.prod(vars[name].numel()), dtype=bool
                )
                for name, is_discrete in self._discrete.items()
            ]
        )

    @invalidate_cache(discrete)
    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
    ) -> SymType:
        """Adds a variable to the NLP problem.

        Parameters
        ----------
        name : str
            Name of the new variable. Must not be already in use.
        shape : tuple[int, int], optional
            Shape of the new parameter. By default a scalar, i.e., ``(1, 1)``.
        discrete : bool, optional
            Flag indicating if the variable is discrete. Defaults to ``False``.

        Returns
        -------
        var : casadi.SX
            The symbol of the new variable.

        Raises
        ------
        ValueError
            Raises if there is already another variable with the same name ``name``.
        """
        if name in self._vars:
            raise ValueError(f"Variable name '{name}' already exists.")
        var = self._sym_type.sym(name, *shape)
        self._vars[name] = var
        self._x = cs.veccat(self._x, var)
        self._has_discrete |= discrete
        self._discrete[name] = discrete
        return var



"""A collection of stand-alone functions that implement some of the basic mathematical
operations that are not available in CasADi. The implementations are simple and thus
not optimized for performance. They are meant to be used as a fallback when the CasADi
really does not provide the required functionality.
"""

from math import sqrt as _sqrt
from typing import Literal, Union

import casadi as cs

SQRT2 = _sqrt(2)


def log(
    x: Union[cs.SX, cs.MX, cs.DM], base: Union[None, cs.SX, cs.MX, cs.DM] = None
) -> Union[cs.SX, cs.MX, cs.DM]:
    r"""Computes the logarithm. With one argument, return the natural logarithm of
    :math:`x` (to base :math:`e`). With two arguments, return the logarithm of :math:`x`
    to the given base :math:`b`, calculated as :math:`\frac{\log x}{\log b}`.

    Parameters
    ----------
    x : casadi.SX, MX or DM
        Value to compute the logarithm of.
    base : casadi.SX, MX or DM, optional
        Base of the logarithm, by default ``None``, so that the natural logarithm is
        computed.

    Returns
    -------
    casadi.SX, MX or DM
        The logarithm of ``x`` with base ``base``, i.e., :math:`\log_{base} x`.
    """
    if base is None:
        return cs.log(x)
    if base == 10 or base == 10.0:
        return cs.log10(x)
    return cs.log(x) / cs.log(base)


def prod(
    x: Union[cs.SX, cs.MX, cs.DM], axis: Literal[0, 1, None, -1, -2] = None
) -> Union[cs.SX, cs.MX, cs.DM]:
    r"""Computes the product of all the elements in ``x`` (CasADi version of
    :func:`numpy.prod`).

    Parameters
    ----------
    x : casadi.SX, MX or DM
        The variable whose entries must be multiplied together.
    axis : {0, 1, None, -1, -2}
        Axis or along which a product is performed. The default, ``axis=None``, will
        calculate the product of all the elements in the matrix. If axis is negative it
        counts from the last to the first axis.

    Returns
    -------
    casadi.SX, MX or DM
        Product of the elements, i.e., :math:`\prod_{i=1}^{|x|}{x_i}`.
    """
    if axis is None:
        x = cs.vec(x)
        axis = 0

    # prod of vector elements
    if x.is_vector():
        if x.shape[axis] == 1:
            return x
        if not isinstance(x, cs.MX):
            return cs.det(cs.diag(x))  # does not work with MX
        p = x[0]
        for i in range(1, x.shape[axis]):
            p *= x[i]
        return p

    # prod of matrix elements
    sum_ = cs.sum1 if (axis == 0 or axis == -2) else cs.sum2
    n_negatives = sum_(x < 0)
    p = cs.exp(sum_(cs.log(cs.fabs(x))))
    return cs.if_else(cs.remainder(n_negatives, 2) == 0.0, 1, -1) * p


def norm_cdf(
    x: Union[cs.SX, cs.MX, cs.DM],
    loc: Union[cs.SX, cs.MX, cs.DM] = 0,
    scale: Union[cs.SX, cs.MX, cs.DM] = 1,
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Computes the cdf of a normal distribution (CasADi version of
    :data:`scipy.stats.norm`'s ``cdf`` method).

    Parameters
    ----------
    x : casadi.SX, MX or DM
        The value at which the cdf is computed.
    loc : casadi.SX, MX or DM, optional
        Mean of the normal distribution. By default, ``loc=0``.
    scale : casadi.SX, MX or DM, optional
        Standard deviation of the normal distribution. By default, ``scale=0``.

    Returns
    -------
    casadi.SX, MX or DM
        The cdf of the normal distribution.
    """
    return 0.5 * (1 + cs.erf((x - loc) / SQRT2 / scale))


def norm_ppf(
    p: Union[cs.SX, cs.MX, cs.DM],
    loc: Union[cs.SX, cs.MX, cs.DM] = 0,
    scale: Union[cs.SX, cs.MX, cs.DM] = 1,
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Computes the quantile (invese of :func:`norm_cdf`) of a normal distribution
    (CasADi version of :data:`scipy.stats.norm`'s ``ppf`` method).

    Parameters
    ----------
    x : casadi.SX, MX or DM
        The value at which the quantile is computed.
    loc : casadi.SX, MX or DM, optional
        Mean of the normal distribution. By default, ``loc=0``.
    scale : casadi.SX, MX or DM, optional
        Standard deviation of the normal distribution. By default, ``scale=0``.

    Returns
    -------
    casadi.SX, MX or DM
        The quantile of the normal distribution.
    """
    return SQRT2 * scale * cs.erfinv(2 * p - 1) + loc


def repeat(
    a: Union[cs.SX, cs.MX, cs.DM], repeats: Union[int, tuple[int, int]]
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Repeats elements in array.

    Parameters
    ----------
    a : casadi.SX or MX or DM
        The array/matrix whose elements are to be repeated.
    repeats : int or tuple of 2 ints
        The number of repeats, in the first and/or second axis.

    Returns
    -------
    casadi.SX, MX or DM
        Output array with repeated elements.
    """
    return cs.kron(a, cs.GenDM_ones(repeats))



from functools import cache, cached_property
from typing import Callable, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.cache import invalidate_cache
from ..core.data import array2cs, cs2array, find_index_in_vector
from ..core.derivatives import hohessian, hojacobian
from ..core.solutions import Solution
from .wrapper import Nlp, Wrapper

SymType = TypeVar("SymType", cs.SX, cs.MX)


class NlpSensitivity(Wrapper[SymType]):
    """Wraps an instance of :class:`csnlp.Nlp` to allow to perform symbolical and
    numerical sensitivity analysis and compute its derivates w.r.t. to different
    quantities, including primal variables and parameters. See
    :cite:`buskens_sensitivity_2001` for nonlinear programming sensitivity analysis. The
    computations are tailored to the IPOPT solver, so no guarantees are provided for
    other solvers. This is due to the fact that IPOPT, as an interior-point solver,
    guarantees that the KKT conditions at the solution are satisfied. This may not the
    case for others.

    Parameters
    ----------
    nlp : Nlp
        The NLP problem to be wrapped.
    target_parameters : casadi.SX or MX, optional
        If provided, computes the sensitibity only w.r.t. these parameters. Of course,
        these parameters must be a subset of all of the NLP's parameters. If ``None``,
        then the derivatives and sensitivity are computed w.r.t. all parameters. If new
        parameters are added after wrapping the nlp in this, be sure to call
        :meth:`set_target_parameters`` again.
    include_barrier_term : bool, optional
        If ``True``, includes in the KKT matrix a new symbolic variable that represents
        the barrier function of the interior-point solver. Otherwise, no additional
        variable is added. See property :meth:`kkt` for more details. By default,
        ``True``.
    """

    def __init__(
        self,
        nlp: Nlp[SymType],
        target_parameters: Optional[SymType] = None,
        include_barrier_term: bool = True,
    ) -> None:
        super().__init__(nlp)
        self.include_barrier_term: bool = include_barrier_term
        self.set_target_parameters(target_parameters)
        self._tau = nlp.sym_type.sym("tau") if include_barrier_term else None

    @property
    def target_parameters(self) -> SymType:
        """Gets the parameters of the NLP that are the target of the sensitivity
        wrapper, i.e., derivatives and sensitivities are limited to these parameters."""
        return self._p_idx[0]

    @cached_property
    def lagrangian(self) -> SymType:
        """Gets the Lagrangian of the NLP problem (usually denoted as ``L``)."""
        return (
            self.nlp.f
            + cs.dot(self.nlp.lam_g, self.nlp.g)
            + cs.dot(self.nlp.lam_h, self.nlp.h)
            + cs.dot(self.nlp.lam_lbx, self.nlp.h_lbx)
            + cs.dot(self.nlp.lam_ubx, self.nlp.h_ubx)
        )

    @cached_property
    def kkt(self) -> tuple[SymType, Optional[SymType]]:
        r"""Gets the KKT conditions of the NLP problem in vector form, i.e.,

        .. math::
            K = \begin{bmatrix}
                \frac{\partial L}{\partial x} \\
                G \\
                \text{diag}(\lambda_h) \cdot H \\
            \end{bmatrix}

        where :math:`\frac{\partial L}{\partial x}` is the gradient of the lagrangian
        w.r.t. the primal variables :math:`x`, :math:`G` collects the equality
        constraints, :math:`H` collects the inequality constraints and :math:`\lambda_h`
        its corresponding dual variables.

        If ``include_barrier_term=True``, the inequalities include an additional barrier
        term ``tau``, so that

        .. math::
            \text{diag}(\lambda_h) \cdot H + \tau

        which is also returned as the second element of the tuple. Otherwise, ``tau`` is
        ``None``.
        """
        tau = self._tau if self._tau is not None else 0
        kkt = cs.vertcat(
            cs.jacobian(self.lagrangian, self.nlp.x).T,
            self.nlp.g,
            (self.nlp.lam_h * self.nlp.h) + tau,
            (self.nlp.h_lbx * self.nlp.lam_lbx) + tau,
            (self.nlp.h_ubx * self.nlp.lam_ubx) + tau,
        )
        return kkt, self._tau

    @cache
    def jacobian(
        self, which: Literal["L-x", "L-p", "g-x", "h-x", "K-p", "K-y"]
    ) -> SymType:
        """Computes a jacobian of one of the NLP problem's quantity w.r.t. another one.

        Parameters
        ----------
        which : {"L-x", "L-p", "g-x", "h-x", "K-p", "K-y"}
            Indicates which jacobian to compute. The possible values are:

            - ``"L-x"``: jacobian of the lagrangian w.r.t. primal variables
            - ``"L-p"``: jacobian of the lagrangian w.r.t. parameters
            - ``"g-x"``: jacobian of the equality constraints w.r.t. primal variables
            - ``"h-x"``: jacobian of the inequality constraints w.r.t. primal variables
            - ``"K-p"``: jacobian of the kkt conditions w.r.t. parameters
            - ``"K-y"``: jacobian of the kkt conditions w.r.t. primal-dual variables.

        Returns
        -------
        SymType
            The requested jacobian, in the form of a symbolic variable.
        """
        if which == "L-x":
            return cs.jacobian(self.lagrangian, self.nlp.x).T
        if which == "L-p":
            p, p_idx = self._p_idx
            return cs.jacobian(self.lagrangian, p)[:, p_idx].T
        if which == "g-x":
            return cs.jacobian(self.nlp.g, self.nlp.x)
        if which == "h-x":
            return cs.jacobian(self.nlp.h, self.nlp.x)
        if which == "K-p":
            p, p_idx = self._p_idx
            return cs.jacobian(self.kkt[0], p)[:, p_idx]
        if which == "K-y":
            return cs.jacobian(self.kkt[0], self.nlp.primal_dual)

    @cache
    def hessian(self, which: Literal["L-pp", "L-xx", "L-px"]) -> SymType:
        """Computes a hessian of the NLP problem's Lagrangian w.r.t. the primal
        variables, the parameters, or both.

        Parameters
        ----------
        which : {"L-pp", "L-xx", "L-px"}
            Indicates which hessian to compute. The possible values are:

            - ``"L-pp"``: hessian of the lagrangian w.r.t. parameters (twice)
            - ``"L-xx"``: hessian of the lagrangian w.r.t. primal variables (twice)
            - ``"L-px"``: hessian of the lagrangian w.r.t. parameters and then primal
              variables

        Returns
        -------
        SymType
            The requested hessian, in the form of a symbolic variable.
        """
        if which == "L-pp":
            p, p_idx = self._p_idx
            return cs.jacobian(self.jacobian("L-p"), p)[:, p_idx]
        if which == "L-xx":
            return cs.jacobian(self.jacobian("L-x"), self.nlp.x)
        if which == "L-px":
            return cs.jacobian(self.jacobian("L-p"), self.nlp.x)

    @cache
    def hojacobian(self, which: Literal["K-pp", "K-yp", "K-yy", "K-py"]) -> np.ndarray:
        """Computes a higher-order jacobian of NLP problem's KKT conditions w.r.t. the
        primal-dual variables, the parameters, or both. See
        :func:`csnlp.core.derivatives.hojacobian` for more information on higher-order
        jacobians.

        Parameters
        ----------
        which : {"K-pp", "K-yp", "K-yy", "K-py"}
            Indicates which hessian to compute. The possible values are:

            - ``"K-pp"``: higher-order jacobian of the KKT conditions w.r.t. parameters
              (twice)
            - ``"K-yp"``: higher-order jacobian of the KKT conditions w.r.t. primal-dual
              variables and then parameters
            - ``"K-yy"``: higher-order jacobian of the KKT conditions w.r.t. primal-dual
              variables (twice)
            - ``"K-py"``: higher-order jacobian of the KKT conditions w.r.t. parameters
              and then primal-dual variables

        Returns
        -------
        SymType
            The requested higher-order jacobian, in the form of a symbolic variable.
        """
        if which == "K-pp":
            p, p_idx = self._p_idx
            return hojacobian(self.jacobian("K-p"), p)[..., p_idx, 0]
        if which == "K-yp":
            p, p_idx = self._p_idx
            return hojacobian(self.jacobian("K-y"), p)[..., p_idx, 0]
        if which == "K-yy":
            return hojacobian(self.jacobian("K-y"), self.nlp.primal_dual)[..., 0]
        if which == "K-py":
            return hojacobian(self.jacobian("K-p"), self.nlp.primal_dual)[..., 0]

    @property
    def licq(self) -> SymType:
        r"""Gets the symbolic matrix for the LICQ, defined as

        .. math::
            LICQ = \begin{bmatrix}
                \frac{\partial g}{\partial x} \\
                \frac{\partial h}{\partial x} \\
                \frac{\partial h_{lbx}}{\partial x} \\
                \frac{\partial h_{ubx}}{\partial x}
            \end{bmatrix}

        If the matrix is linear independent, then the NLP satisfies the Linear
        Independence Constraint Qualification.

        Notes
        -----
        Theoretically, the LICQ is computed for only the active inenquality constraints.
        Since this is merely a symbolic representation, all constraints are included,
        and it's up to the user to eliminate the inactive ones.
        """
        return cs.vertcat(
            self.jacobian("g-x"),
            self.jacobian("h-x"),
            cs.jacobian(self.nlp.h_lbx, self.nlp.x),
            cs.jacobian(self.nlp.h_ubx, self.nlp.x),
        )

    def parametric_sensitivity(
        self,
        expr: SymType = None,
        solution: Optional[Solution[SymType]] = None,
        second_order: bool = False,
    ) -> Union[tuple[SymType, Optional[SymType]], tuple[cs.DM, Optional[cs.DM]]]:
        r"""Performs the (symbolic or numerical) sensitivity of the NLP w.r.t. its
        parametrization, according to :cite:`buskens_sensitivity_2001`.

        Parameters
        ----------
        expr : casadi.SX or MX, optional
            If provided, computes the sensitivity of this expression (which must be
            dependent on the primal-dual variables and/or parameters of the NLP) w.r.t.
            the NLP parameters. If ``None``, then the sensitivity of the primal-dual
            variables is returned.
        solution : Solution, optional
            If a solution is passed, then the sensitivity is numerically computed for
            that solution; otherwise, the sensitivity is carried out symbolically
            (however, this is much more computationally intensive).
        second_order : bool, optional
            If ``second_order=False``, the analysis is stopped at the first order;
            otherwise, also second-order information is computed.

        Returns
        -------
        2-element tuple of casadi.SX, or MX, or DM
            The 1st and 2nd-order NLP parametric sensitivity in the form of an array/DM,
            if a solution is passed, or a symbolic vector SX/MX. When
            ``second_order=False`` the second element in the tuple is ``None``.

        Raises
        ------
        numpy.linalg.LinAlgError
            Raises if the KKT conditions lead to a singular matrix.
        """
        # first and second order sensitivities, a.k.a., dydp and d2dydp2
        d: Callable[[SymType], Union[SymType, cs.DM]] = (
            (lambda o: o)
            if solution is None
            else (lambda o: solution.value(o))  # type: ignore[union-attr]
        )
        dydp, dydp_np, d2ydp2 = self._y_parametric_sensitivity(
            solution, second_order, d
        )

        # first order sensitivity of Z, a.k.a., dZdp
        Z = expr  # Z := z(x(p),lam(p),p)
        if Z is not None:
            Zshape = Z.shape
            Z = cs.vec(Z)
            y = self.nlp.primal_dual
            p, p_idx = self._p_idx
            np_ = p.shape[0] if isinstance(p_idx, slice) else p_idx.size

            if second_order:
                Zpp, Zp = hohessian(Z, p)
                Zyy, Zy = hohessian(Z, y)
            else:
                Zp = hojacobian(Z, p)
                Zy = hojacobian(Z, y)
            Zp = d(array2cs(Zp[:, 0, p_idx, 0]))
            Zy = d(array2cs(Zy[:, 0, :, 0]))

            dZdp = Zy @ dydp + Zp
            dZdp = cs2array(dZdp).reshape(Zshape + (np_,), order="F").squeeze()
            if dZdp.ndim <= 2:
                dZdp = array2cs(dZdp)

        # if no second order is required, just return
        if not second_order:
            return dydp if Z is None else dZdp, None

        # if no expression is given, just return
        if Z is None:
            d2ydp2 = d2ydp2.squeeze()
            if d2ydp2.ndim <= 2:
                d2ydp2 = array2cs(d2ydp2)
            return dydp, d2ydp2

        # second order sensitivity of Z, a.k.a., d2Zdp2
        Zyp = d(hohessian(Z, y, p)[0][:, 0, :, 0][:, :, p_idx, 0])
        Zpy = d(hohessian(Z, p, y)[0][:, 0, p_idx, 0][:, :, :, 0])
        Zpp = d(Zpp[:, 0, p_idx, 0][:, :, p_idx, 0])
        Zyy = d(Zyy[:, 0, :, 0][:, :, :, 0])
        T1 = (d2ydp2.transpose((1, 2, 0)) @ cs2array(Zy).T).transpose((2, 0, 1))
        T2 = Zyy.transpose((0, 2, 1)) @ dydp_np + Zpy.transpose((0, 2, 1)) + Zyp
        d2Zdp2 = Zpp + T1 + dydp_np.T @ T2
        d2Zdp2 = d2Zdp2.reshape(Zshape + (np_, np_), order="F").squeeze()
        if d2Zdp2.ndim <= 2:
            d2Zdp2 = array2cs(d2Zdp2)
        return dZdp, d2Zdp2

    def _y_parametric_sensitivity(
        self,
        solution: Optional[Solution[SymType]],
        second_order: bool,
        d: Callable[[SymType], Union[SymType, cs.DM]],
    ) -> Union[tuple[SymType, np.ndarray, SymType], tuple[cs.DM, np.ndarray, cs.DM]]:
        """Internal utility to compute the sensitivity of ``y`` w.r.t. ``p``."""
        # first order sensitivity, a.k.a., dydp
        Ky = d(self.jacobian("K-y"))
        Kp = d(self.jacobian("K-p"))
        dydp = (cs.solve if solution is None else np.linalg.solve)(-Ky, Kp)
        if not second_order:
            return dydp, None, None  # type: ignore[return-value]

        # second order sensitivity, a.k.a., d2ydp2
        dydp_ = cs2array(dydp)
        Kpp = d(self.hojacobian("K-pp"))
        Kpy = d(self.hojacobian("K-py"))
        Kyp = d(self.hojacobian("K-yp"))
        Kyy = d(self.hojacobian("K-yy"))
        M = (
            Kpp
            + (Kpy.transpose((0, 2, 1)) + Kyp + (Kyy @ dydp_)).transpose((0, 2, 1))
            @ dydp_
        ).transpose((2, 0, 1))
        if solution is None:
            d2ydp2 = -(cs2array(cs.inv(Ky)) @ M).transpose((1, 2, 0))
        else:
            d2ydp2 = -np.linalg.solve(Ky, M).transpose((1, 2, 0))
        return dydp, dydp_, d2ydp2

    @invalidate_cache(jacobian, hessian, hojacobian)
    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        """See :meth:`csnlp.Nlp.parameter`."""
        return self.nlp.parameter(name, shape)

    @invalidate_cache(lagrangian, kkt, jacobian, hessian, hojacobian)
    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        """See :meth:`csnlp.Nlp.variable`."""
        return self.nlp.variable(name, shape, discrete, lb, ub)

    @invalidate_cache(lagrangian, kkt, jacobian, hessian, hojacobian)
    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        """See :meth:`csnlp.Nlp.constraint`."""
        return self.nlp.constraint(name, lhs, op, rhs, soft, simplify)

    @invalidate_cache(lagrangian, kkt, jacobian, hessian, hojacobian)
    def minimize(self, objective: SymType) -> None:
        """See :meth:`csnlp.Nlp.minimize`."""
        return self.nlp.minimize(objective)

    @invalidate_cache(jacobian, hessian, hojacobian)
    def set_target_parameters(self, parameters: Optional[SymType]) -> None:
        """Sets the target parameters of the sensitivity wrapper.

        Parameters
        ----------
        parameters : casadi.MX or SX or None
            New parameters to target during the sensitivity analyses. If ``None``, all
            NLP parameters are included.
        """
        if parameters is None:
            self._p_idx_internal = None, slice(None)
            return
        p: SymType = cs.vec(parameters)
        if self.nlp.sym_type is cs.SX:
            self._p_idx_internal = p, slice(None)
            return
        self._p_idx_internal = self.nlp.p, find_index_in_vector(self.nlp.p, p)

    @property
    def _p_idx(self) -> tuple[SymType, Union[slice, npt.NDArray[np.int64]]]:
        """Internal utility to return the indices of ``p`` from all the NLP pars. While
        SX is fine with computing jacobians with indexed variables, MX requires purely
        symbolic variables. So, for MX, jacobians need to be computed for all elements
        and then indexed."""
        p = self._p_idx_internal[0]
        return (self.nlp.p if p is None else p), self._p_idx_internal[1]
